#include <_null.h>
#include <math.h>

class TObject {
    TColor color;
  public:
    TObject(double r, double g, double b)
    {
      next = NULL;
      color.r = r;
      color.g = g;
      color.b = b;
    };
    virtual int Intersect(double v_x, double v_y, double v_z,
                          double c_x, double c_y, double c_z,
                          double &x, double &y, double &z) = 0;
    TColor GetColor() { return color; }
    TObject *next;
    ~TObject()
    {
      if (next) delete next;
    }
};

class THorizPlane: public TObject {
    double height;
  public:
    THorizPlane(double h, double r, double g, double b):
               TObject(r, g, b)
    {
      height = h;
    };
    int Intersect(double v_x, double v_y, double v_z,
                  double c_x, double c_y, double c_z,
                  double &x, double &y, double &z)
    {
      if ((c_z < height && v_z < 0) || (c_z > height && v_z > 0) || !v_z)
        return 0;

      if (c_z == height)
      {
        x = c_x;
        y = c_y;
        z = c_z;
        return 1;
      }

      double t = fabs((c_z - height) / v_z);
      x = c_x + v_x * t;
      y = c_y + v_y * t;
      z = height;

      return 1;
    }
};

class TVertXPlane: public TObject {
    double depth;
  public:
    TVertXPlane(double d, double r, double g, double b):
               TObject(r, g, b)
    {
      depth = d;
    };
    int Intersect(double v_x, double v_y, double v_z,
                  double c_x, double c_y, double c_z,
                  double &x, double &y, double &z)
    {
      if ((c_y < depth && v_y < 0) || (c_y > depth && v_y > 0) || !v_y)
        return 0;

      if (c_y == depth)
      {
        x = c_x;
        y = c_y;
        z = c_z;
        return 1;
      }

      double t = fabs((c_y - depth) / v_y);
      x = c_x + v_x * t;
      y = depth;
      z = c_z + v_z * t;

      return 1;
    }
};

class TVertYPlane: public TObject {
    double side;
  public:
    TVertYPlane(double s, double r, double g, double b):
               TObject(r, g, b)
    {
      side = s;
    };
    int Intersect(double v_x, double v_y, double v_z,
                  double c_x, double c_y, double c_z,
                  double &x, double &y, double &z)
    {
      if ((c_x < side && v_x < 0) || (c_x > side && v_x > 0) || !v_x)
        return 0;

      if (c_x == side)
      {
        x = c_x;
        y = c_y;
        z = c_z;
        return 1;
      }

      double t = fabs((c_x - side) / v_x);
      x = side;
      y = c_y + v_y * t;
      z = c_z + v_z * t;

      return 1;
    }
};
