#include <fstream.h>
#include <math.h>
#include "objects.cpp"
#include "color.h"

#define MAX_LIGHTS  10

struct TLight {
  double x, y, z;
  double amp;
  char active;
};

class TScene {
    int w, h, c_x, c_y;
    double light_x, light_y, light_z;
    double x, y, z, perspect;
    TObject *objects, *end;
    TLight lights[MAX_LIGHTS];
  public:
    TScene(int width, int height, double perspect, int center_x, int center_y,
           double camera_x, double camera_y, double camera_z,
           double view_x, double view_y, double view_z);
    void SetupLight(int number, double x, double y, double z, double amp)
    {
      lights[number].x = x;
      lights[number].y = y;
      lights[number].z = z;
      lights[number].amp = amp;
    }
    void EnableLight(int number)
    {
      lights[number].active = 1;
    }
    void DisableLight(int number)
    {
      lights[number].active = 0;
    }
    void SetCamera(double view_x, double view_y, double view_z)
    {
      x = view_x;
      y = view_y;
      z = view_z;
    }
    virtual void PutPixel(int x, int y, TColor color) = 0;
    void RenderScene();
    TColor AddColors(TColor c1, TColor c2)
    {
      TColor t;

      if ((unsigned)c1.r + (unsigned)c2.r > 255)
        t.r = 255;
      else
        t.r = c1.r + c2.r;

      if ((unsigned)c1.g + (unsigned)c2.g > 255)
        t.g = 255;
      else
        t.g = c1.g + c2.g;

      if ((unsigned)c1.b + (unsigned)c2.b > 255)
        t.b = 255;
      else
        t.b = c1.b + c2.b;

      return t;
    }
    TColor RangeColor(double r, double g, double b)
    {
      TColor t;

      if (r > 255)
        t.r = 255;
      else
        t.r = (unsigned char)r;

      if (g > 255)
        t.g = 255;
      else
        t.g = (unsigned char)g;

      if (b > 255)
        t.b = 255;
      else
        t.b = (unsigned char)b;

      return t;
    }
    void AddObject(TObject *o)
    {
      if (!objects)
        objects = o;
      else
        end -> next = o;
      end = o;
    }
    double F1(double x, double y, double z)
    {
      return 2;
      return sin(y * M_PI / 36) + sin(x * M_PI / 36);
    }
    ~TScene()
    {
      delete objects;
    }
};

TScene::TScene(int width, int height, double perspect, int center_x, int center_y,
               double camera_x, double camera_y, double camera_z,
               double view_x, double view_y, double view_z)
{
  w = width;
  h = height;
  light_x = 0;
  light_y = 0;
  light_z = 0;
  c_x = center_x;
  c_y = center_y;
  TScene::perspect = perspect;
  SetCamera(camera_x, camera_y, camera_z);
  objects = NULL;
  end = NULL;
  for (int i = 0; i < MAX_LIGHTS; i++)
  {
    lights[i].x = 0;
    lights[i].y = 0;
    lights[i].z = 0;
    lights[i].amp = 0;
    lights[i].active = 0;
  }
}

void TScene::RenderScene()
{
  TColor color, obj_color;
  TObject *t;
  double i_x, i_y, i_z, r_x, r_y, r_z;
  double r, g, b, lightning;
  double dist, d;
  TObject *o;

//  ofstream out("room.bmp", ios::binary);

  for (int j = h / 2; j >= -h / 2; j--)
    for (int i = -w / 2; i < w / 2; i++)
    {
      t = objects;
      dist = 1000000;
      r = 0;
      g = 0;
      b = 0;

      while (t)
      {
        if (t -> Intersect(i, perspect, j,
                           x, y, z, i_x, i_y, i_z))
        {
          d = sqrt((i_x - x) * (i_x - x) +
                   (i_y - y) * (i_y - y) +
                   (i_z - z) * (i_z - z));
          if (d < dist)
          {
            dist = d;
            r_x = i_x;
            r_y = i_y;
            r_z = i_z;
            o = t;
          }
        }
        t = t -> next;
      }

      if (dist == 1000000) continue;

      lightning = 0;
      for (int n = 0; n < MAX_LIGHTS; n++)
      {
        if (!lights[n].active) continue;

        d = sqrt((lights[n].x - r_x) * (lights[n].x - r_x) +
                 (lights[n].y - r_y) * (lights[n].y - r_y) +
                 (lights[n].z - r_z) * (lights[n].z - r_z));

        double q = lights[n].amp / d;

        if (q > lightning) lightning = q;
      }

      obj_color = o -> GetColor();
      r = obj_color.r;
      g = obj_color.g;
      b = obj_color.b;
      r *= lightning / 256;
      g *= lightning / 256;
      b *= lightning / 256;
      color = RangeColor(lightning, lightning, lightning);
      color = AddColors(color, RangeColor(r, g, b));

      PutPixel(c_x + i, c_y - j, color);
//      out.put(color.b);
//      out.put(color.g);
//      out.put(color.r);
    }
//      out.close();
}
